/*BINFMTCXX: -Igeometry/include -Iboost_1_76_0
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <iostream>
#include <fstream>
#include <boost/geometry.hpp>
#include <boost/geometry/geometries/point.hpp>
#include <boost/geometry/geometries/box.hpp>
#include <boost/geometry/index/rtree.hpp>
#include <boost/interprocess/managed_mapped_file.hpp>

namespace b = boost;
namespace bg = b::geometry;
namespace bgm = bg::model;
namespace idx = bg::index;
namespace ipc = b::interprocess;

typedef bgm::point <double, 2, bg::cs::cartesian> point_t;
typedef bgm::box<point_t> box_t;
typedef uint64_t pl_id_t;

typedef std::pair<box_t, pl_id_t> value_t;
typedef idx::linear<51, 25> params_t;
typedef idx::indexable<value_t> indexable_t;
typedef idx::equal_to<value_t> equal_to_t;
typedef ipc::allocator<value_t, ipc::managed_mapped_file::segment_manager> allocator_t;
typedef idx::rtree<value_t,
                   params_t,
                   indexable_t,
                   equal_to_t,
                   allocator_t> rtree_t;

typedef ipc::managed_mapped_file mmfile;

using namespace std;

static allocator_t *alloc = NULL;
static rtree_t *tree = NULL;

static const size_t minsize = 10000000;
static size_t current_size = minsize;
static string outfilename;
static mmfile *outmap;

off_t fsize(const char *filename)
{
  struct stat st;

  if (stat(filename, &st) == 0)
    return st.st_size;

  return -1;
}

struct file_bounding_box {
  pl_id_t id;
  double x1;
  double y1;
  double x2;
  double y2;
} __attribute__ ((packed));

class ControlFlow : exception {};

class StreamEnded : ControlFlow {};

std::istream& operator>>(std::istream &is, value_t &v)
{
  file_bounding_box tmp;
  is.read((char *)&tmp, sizeof(file_bounding_box));
  if (is)
    {
      v.second = tmp.id;
      bg::set<bg::min_corner, 0>(v.first, tmp.x1);
      bg::set<bg::min_corner, 1>(v.first, tmp.y1);
      bg::set<bg::max_corner, 0>(v.first, tmp.x2);
      bg::set<bg::max_corner, 1>(v.first, tmp.y2);
      return is;
    }
  else
    {
      if (is.gcount() == 0)
        {
          // All good, we merely reached the end of our input.
          throw StreamEnded();
        }
      else
        {
          throw runtime_error("Corrupt input");
        }
    }
}

static mmfile *make_outmap(size_t size)
{
  return new mmfile(ipc::open_or_create, outfilename.c_str(), size);
}

static void get_tree()
{
  alloc = new allocator_t(outmap->get_segment_manager());

  tree = outmap->find_or_construct<rtree_t>("rtree")            \
    (params_t(), indexable_t(), equal_to_t(), *alloc);
}

static void release_tree_and_map()
{
  tree = NULL;
  if (alloc)
    delete alloc;
  alloc = NULL;
  if (outmap)
    {
      outmap->flush();
      delete outmap;
    }
  outmap = NULL;
}

static void maybe_enlarge_tree(size_t threshold)
{
  if (outmap->get_segment_manager()->get_free_memory() <= threshold)
    {
      release_tree_and_map();
      size_t extrasize = current_size;
      if (mmfile::grow(outfilename.c_str(), extrasize))
        {
          current_size += extrasize;
        }
      else
        {
          cerr << "WARNING: grow() appears to have failed!" << endl;
        }
      outmap = make_outmap(current_size);
      get_tree();
    }
}

static void build_from_stream(istream &in)
{
  value_t tmp;

  get_tree();

  size_t i = 0;

  try
    {
      for(;;)
        {
          try
            {
              in >> tmp;
            }
          catch (std::exception& e)
            {
              cerr << "Possibly corrupt input.  This is what happened: " <<
                e.what() << endl;
              throw e;
            }
          maybe_enlarge_tree(1 << 20); // ~1M
          tree->insert(tmp);
        }
    }
  catch(StreamEnded &e)
    {
      outmap->flush();
    }
}

void print_usage(int argc, char *argv[])
{
  cerr << "Usage: " << argv[0] << " [-i infile] [-s size] outfile" << endl;
  cerr << "\t-h\t\tprint this help message" << endl;
  cerr << "\t-i infile\tuse infile as input; defaults to stdin; " <<
    "'-' means stdin" << endl;
  cerr << "\t-s size\t\tmake the output file size bytes large; " <<
    "defaults to 10MB" << endl;
}

int main(int argc, char *argv[])
{
  string infile = "-";
  char c;

  while ((c = getopt(argc, argv, "hi:s:")) != -1)
    {
      switch (c)
        {
        case 'h':
          print_usage(argc, argv);
          return 0;
        case 'i':
          infile = optarg;
          break;
        case 's':
          {
            try
              {
                istringstream ss(optarg);
                ss >> current_size;
                if (current_size < minsize)
                  {
                    cerr << "WARNING: initial size is less than the default.  "
                         << "Using " << minsize << " bytes as size" << endl;
                    current_size = minsize;
                  }
              }
            catch (...)
              {
                print_usage(argc, argv);
                return 1;
              }
          }
          break;
        case '?':
        default:
          print_usage(argc, argv);
          return 1;
        }
    }
  if (optind + 1 != argc)
    {
      print_usage(argc, argv);
      return 1;
    }
  outfilename = argv[optind];

  try
    {
      outmap = make_outmap(current_size);
    }
  catch (std::exception& e)
    {
      cerr << "Opening output file failed: " << e.what() << endl;
      return 2;
    }

  try
    {
      if (infile == "-")
        {
          build_from_stream(cin);
        }
      else
        {
          ifstream ins;
          ins.open(infile.c_str());
          build_from_stream(ins);
        }
    }
  catch (std::exception& e)
    {
      cerr << "Internal error: " << e.what() << endl;
    }

  release_tree_and_map();
  mmfile::shrink_to_fit(outfilename.c_str());
  return 0;
}
